# -*- coding: utf-8 -*-
"""
Module providing an entry point to DID/identifier operations via a CLI.
"""

# Created: 2019-05-08 Guy K. Kloss <guy@mysinglesource.io>
#
# (c) 2018-2019 by SingleSource Limited, Auckland, New Zealand
#     http://mysinglesource.io/
#     Apache 2.0 Licence.
#
# This work is licensed under the Apache 2.0 open source licence.
# Terms and conditions apply.
#
# You should have received a copy of the licence along with this
# program.

__author__ = 'Guy K. Kloss <guy@mysinglesource.io>'

import argparse
import base64
import logging
from typing import Optional

from singlesource.dids import (generate_id,
                               make_did_document,
                               api_register_did_document)


logger = logging.getLogger(__name__)


def generate_did(identifier: Optional[str] = None,
                 pub_key: Optional[bytes] = None,
                 id_type: str = 'did'):
    """
    Generate a (decentralised) identifier from a unique identifier.

    :param identifier: The specific identifier to use, usually a blockchain
        address (e.g. an encoded public key).
    :param pub_key: Base64 encoded public signing key. Currently only supported
        for Ed25519 keys.
    :param id_type: Identifier type, defaulting to a DID (decentralised
        identifier).
    """
    if identifier:
        print(generate_id(identifier=identifier, id_type=id_type))
    elif pub_key:
        print(generate_id(public_key=base64.b64decode(pub_key),
                          id_type=id_type))
    else:
        raise ValueError('Either an identifier or a public key is required'
                         ' to generate a DID.')


def generate_did_document_from_signing_key(did: str, pub_key: str,
                                           did_doc_file: str):
    """
    Generate a DID document from a public (Ed25519) signing key.

    :param did: Decentralised identifier.
    :param pub_key: Base64 encoded public signing key. Currently only supported
        for Ed25519 keys.
    :param did_doc_file: File to write the DID document to.
    """
    public_keys = [
        {'key_type': 'Ed25519VerificationKey2018',
         'public_key': base64.b64decode(pub_key),
         'authentication_type': 'Ed25519SignatureAuthentication2018'}
    ]
    with open(did_doc_file, 'wt') as fd:
        fd.write(make_did_document(did, public_keys))


def register_did_document_api(did: str, did_doc_file: str):
    """
    Register a DID document with the DID registry (API).

    This is an interim (centralised) solution using a centralised DID registry
    implemented via a REST API.

    :param did: Decentralised identifier.
    :param did_doc_file: File containing the DID document for registration.
    """
    with open(did_doc_file, 'rt') as fd:
        api_register_did_document(did, fd.read())


def main(args: argparse.Namespace):
    """
    Delegate to the right functions from given (command line) arguments.

    :param args: Command line arguments provided:

        - args.id_type - type: str
        - args.address - type: str
        - args.did - type: str
        - args.pub_sig_key - type: str
        - args.did_document - type: str
    """
    if args.operation == 'generate':
        if args.address:
            generate_did(identifier=args.address, id_type=args.id_type)
        elif args.pub_sig_key:
            generate_did(pub_key=args.pub_sig_key, id_type=args.id_type)
        else:
            raise ValueError('Either an identifier or a public key is required'
                             ' to generate a DID.')
    elif args.operation == 'makedoc':
        generate_did_document_from_signing_key(args.did, args.pub_sig_key,
                                               args.did_document)
    elif args.operation == 'register':
        register_did_document_api(args.did, args.did_document)
    elif args.operation == 'resolve':
        raise NotImplementedError('DID operation not implemented, yet: {}'
                                  .format(args.operation))
    else:
        raise ValueError('Unsupported operation "{}" for DIDs.'
                         .format(args.operation))

    logger.info('Operation "{}" finished.'.format(args.operation))



# -*- coding: utf-8 -*-
"""High level (console) scripts."""

# Created: 2019-05-08 Guy K. Kloss <guy@mysinglesource.io>
#
# (c) 2018-2019 by SingleSource Limited, Auckland, New Zealand
#     http://mysinglesource.io/
#     Apache 2.0 Licence.
#
# This work is licensed under the Apache 2.0 open source licence.
# Terms and conditions apply.
#
# You should have received a copy of the licence along with this
# program.



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Command line (CLI) client for SingleSource operations.
"""

# Created: 2019-05-08 Guy K. Kloss <guy@mysinglesource.io>
#
# (c) 2018-2019 by SingleSource Limited, Auckland, New Zealand
#     http://mysinglesource.io/
#     Apache 2.0 Licence.
#
# This work is licensed under the Apache 2.0 open source licence.
# Terms and conditions apply.
#
# You should have received a copy of the licence along with this
# program.

__author__ = 'Guy K. Kloss <guy@mysinglesource.io>'

import argparse
import logging

from sspyjose import Jose

from singlesource.config import config
from singlesource.scripts import (dids,
                                  attestations)


logger = logging.getLogger(__name__)

# Configure to use AES256-GCM as a default cipher for encryption.
# Current options: 'C20P' (ChaCha20/Poly1305), 'A256GCM' (AES256-GCM).
Jose.DEFAULT_ENC = config.default_enc


def _console():
    """
    Entry point for console execution.
    """
    # Set up logger.
    logging.basicConfig(
        level=logging.INFO,
        format='%(levelname)s\t%(name)s\t%(asctime)s %(message)s')

    # Create the top-level parser.
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(help='sub-commands')

    # Create the parser for the "claims" command.
    parser_dids = subparsers.add_parser(
        'dids', help='DID- or identifier-related operations')
    parser_dids.set_defaults(func=dids.main)
    parser_dids.add_argument(
        '--operation', type=str,
        help='Operation to perform, one of:'
             ' `generate`, `makedoc`, `register`, `resolve`.')
    parser_dids.add_argument(
        '--type', dest='id_type', type=str, default='did',
        help="Identifier type, one of: `did` (default), `imid`.")
    parser_dids.add_argument(
        '--address', type=str, default=None,
        help='Public address of identity (e.g. blockchain address).')
    parser_dids.add_argument(
        '--did', type=str, default=None,
        help='DID or public identifier of identity.')
    parser_dids.add_argument(
        '--did-doc', dest='did_document', type=str, default=None,
        help='File name for the DID document.')
    parser_dids.add_argument(
        '--pub-sig-key', dest='pub_sig_key', type=str, default=None,
        help='Base64 encoded public signing key (Ed25519).')

    # Create the parser for the "claims" command.
    parser_attestations = subparsers.add_parser(
        'attestations', help='Attestation-related operations')
    parser_attestations.set_defaults(func=attestations.main)
    parser_attestations.add_argument(
        '--operation', type=str,
        help='Operation to perform, one of:'
             ' `attest`, `xxx`.')
    parser_attestations.add_argument(
        '--subject', type=str, default=None,
        help="DID or public identifier of subject's identity.")
    parser_attestations.add_argument(
        '--attester-obo', dest='attester_obo', type=str, default=None,
        help='DID of the verifying party requesting the attestation'
             ' (default: from configuration).')
    parser_attestations.add_argument(
        '--statements', type=argparse.FileType('rt'), default=None,
        help='Attestation statements to use (default: from configuration).')
    parser_attestations.add_argument(
        '--claims', type=argparse.FileType('rt'), default=None,
        help='File for the claims.')
    parser_attestations.add_argument(
        '--attestation', type=str, default=None,
        help='File for the attestations.')

    # Now let's do the work.
    args = parser.parse_args()
    args.func(args)


if __name__ == '__main__':
    _console()



# -*- coding: utf-8 -*-
"""
Module providing an entry point to attestation operations via a CLI.
"""

# Created: 2019-05-09 Guy K. Kloss <guy@mysinglesource.io>
#
# (c) 2018-2019 by SingleSource Limited, Auckland, New Zealand
#     http://mysinglesource.io/
#     Apache 2.0 Licence.
#
# This work is licensed under the Apache 2.0 open source licence.
# Terms and conditions apply.
#
# You should have received a copy of the licence along with this
# program.

__author__ = 'Guy K. Kloss <guy@mysinglesource.io>'

import argparse
import io
import json
import logging
from typing import Optional

from singlesource.attestations import api_request_kym_attestation
from singlesource.config import config


logger = logging.getLogger(__name__)


def attest_kym_claim_set(subject: str,
                         claims_fd: io.TextIOBase,
                         attestation_file: str,
                         attester_obo: Optional[str] = None,
                         statements_fd: Optional[io.TextIOBase] = None):
    """
    Attest a claim set for a device using the KYM API.

    :param subject: DID of the claim set subject.
    :param claims_fd: File for the claims.
    :param attestation_file: File name to store the attestation in.
    :param attester_obo: DID of the organisation/identity on whose behalf
        the attestation will be requested.
    :param statements_fd: Name of JSON file containing attestation statements.
    """
    plain_claims = json.load(claims_fd)
    statements = config.kym_default_statements
    if statements_fd:
        statements = json.load(statements_fd)
    attester_obo = attester_obo or config.kym_default_attester_obo
    result = api_request_kym_attestation(subject, plain_claims,
                                         attester_obo=attester_obo,
                                         statements=statements,
                                         device=True)
    with open(attestation_file, 'wt') as fd:
        json.dump(result, fd, indent=2)


def main(args: argparse.Namespace):
    """
    Delegate to the right functions from given (command line) arguments.

    :param args: Command line arguments provided:

        - args.subject - type: str
        - args.claims - type: io.TextIOBase
        - args.attester_obo - type: str
        - args.statements - type: io.TextIOBase
        - args.attestation - type: str
    """
    if args.operation == 'attest':
        attest_kym_claim_set(args.subject, args.claims, args.attestation,
                             args.attester_obo, args.statements)
    else:
        raise ValueError('Unsupported operation "{}" for attestations.'
                         .format(args.operation))

    logger.info('Operation "{}" finished.'.format(args.operation))



# -*- coding: utf-8 -*-
"""
Configuration settings for config files and defaults.

System and user level configuration files are detected and automatically
read. The configuration files are following a YAML syntax.
"""

# Created: 2019-03-15 Guy K. Kloss <guy@mysinglesource.io>
#
# (c) 2018-2019 by SingleSource Limited, Auckland, New Zealand
#     http://mysinglesource.io/
#     Apache 2.0 Licence.
#
# This work is licensed under the Apache 2.0 open source licence.
# Terms and conditions apply.
#
# You should have received a copy of the licence along with this
# program.

__author__ = 'Guy K. Kloss <guy@mysinglesource.io>'

import logging
import os
import yaml

import appdirs
import sspyjose


class _Settings(object):
    """Default configuration options."""

    # ### General options.
    # Default time to live span of claims and attestations made.
    default_ttl = 365 * 24 * 3600  # 1 year in seconds.

    # ### Logging options.
    # Log level.
    log_level = logging.INFO
    # Log file name.
    log_file = '/var/log/pysinglesource/pysinglesource.log'
    # Maximal number of bytes per log.
    log_max_bytes = 2 * 1024 ** 2  # 2 MiB
    # Number of log backups for log rotation.
    log_backup_count = 5
    # Default cipher for encryption (`alg` value for JWE).
    # Current options: 'C20P' (ChaCha20/Poly1305), 'A256GCM' (AES256-GCM).
    default_enc = 'A256GCM'

    # ### Credentials.
    # OAuth client credentials.
    oauth_url = 'https://auth.mysinglesource.io/kym/login'
    oauth_consumer_key = None  # type: str
    oauth_consumer_secret = None  # type: str

    # ### API end-points.
    # DID registry REST API.
    did_registry_base_url = 'https://dev.mysinglesource.io'
    did_registry_resource_path = '/diddocuments'
    # KYM attestation REST API.
    kym_attestation_base_url = 'https://dev.mysinglesource.io'
    kym_attestation_resource_path = '/kymdevices'

    # ### KYM attestation defaults.
    # Default statements template for KYM devices.
    kym_default_statements = [
        {
            "ttl": 315360000,  # 10 years
            "metadata": {
                "kym_compliance": "device identity assurance, level 3",
                "governance": "IoT association"
            }
        }
    ]
    # Default entity/organisation on whose behalf the attestation will be made.
    kym_default_attester_obo = (
        'did:ssid:5E2J5Uqaf73CoKyD6itv4FcQvgVZ7nZR9rmmS3cCGbYNxEv7')

    def __init__(self, appname: str):
        """
        Initialise the configuration.

        Defaults are overridden with system and/or user configurations.
        """
        config_paths = [appdirs.user_config_dir(appname),
                        appdirs.site_config_dir(appname)]

        for config_path in config_paths:
            if os.path.exists(config_path):
                file_config = yaml.load(open(config_path))
                for key, value in file_config.items():
                    setattr(self, key, value)

        sspyjose.Jose.DEFAULT_ENC = self.default_enc


config = _Settings('pysinglesource')



# -*- coding: utf-8 -*-
"""DID handling support."""

# Created: 2019-05-08 Guy K. Kloss <guy@mysinglesource.io>
#
# (c) 2018-2019 by SingleSource Limited, Auckland, New Zealand
#     http://mysinglesource.io/
#     Apache 2.0 Licence.
#
# This work is licensed under the Apache 2.0 open source licence.
# Terms and conditions apply.
#
# You should have received a copy of the licence along with this
# program.

import base64
import json
import nacl.hash
from typing import Optional, Iterable, Dict

import base58
import nacl.encoding
import nacl.utils

from singlesource import utils
from singlesource.config import config  # noqa: F401 # @UnusedImport


# To load the default configuration.
_IDENTIFIER_FORMAT = '{}:ssid:{}'
_IMID_SIZE = 32
_ADDRESS_TYPE = b'*'  # 42: CENNZnet address type.
_ADDRESS_TYPE_IMID = b'?'  # 63: Last unreserved of Substrate.
_PREFIX = b'SS58PRE'
_CHECKSUM_SIZE = 2
_DIGEST_LENGTH = 64
_DID_DOC_SKELETON = {
    '@context': 'https://w3id.org/did/v1',
    'id': None,
    'publicKey': [],
    'authentication': []
}


def generate_id(identifier: Optional[str] = None,
                public_key: Optional[bytes] = None,
                id_type: str = 'did') -> str:
    """
    Generate a (decentralised) identifier from a unique identifier.

    :param identifier: The specific identifier to use, usually a blockchain
        address (e.g. an encoded public key).
    :param public_key: The public key to use for generating an identifier.
    :param id_type: Identifier type, defaulting to a DID (decentralised
        identifier).
    :return: Identifier representation usable in the decentralised environment.
    """
    if id_type not in ('did', 'imid'):
        raise ValueError('Identifier type unknow/unsupported: {}'
                         .format(id_type))
    if not identifier:
        if id_type == 'did':
            identifier = _get_address_from_pub_key(public_key)
        elif id_type == 'imid':
            imid_raw = nacl.utils.random(_IMID_SIZE)
            identifier = _get_address_from_pub_key(imid_raw, id_type=id_type)
    return _IDENTIFIER_FORMAT.format(id_type, identifier)


def _get_address_from_pub_key(pub_key: bytes, id_type: str = 'did') -> str:
    """
    Generate an address (blockchain address) from the public key.

    TODO: This needs potential fixing, as the checksum doesn't pan out.

    :param pub_key: Public key or identifier bytes to use for generation.
    :param id_type: Identifier type, defaulting to a DID (decentralised
        identifier).
    :return: Identifier representation usable in the decentralised environment.
    """
    address_type = _ADDRESS_TYPE if id_type == 'did' else _ADDRESS_TYPE_IMID
    payload = address_type + pub_key
    checksum = nacl.hash.blake2b(_PREFIX + payload, digest_size=_DIGEST_LENGTH,
                                 encoder=nacl.encoding.RawEncoder)
    complete = payload + checksum[:_CHECKSUM_SIZE]
    return base58.b58encode(complete).decode('utf-8')


def make_did_document(did: str, public_keys: Iterable[Dict]) -> str:
    """
    Generate a DID document from a DID and public keys.

    This is to suffice the needs of the `api_register_did_document()` function
    in this module.

    :param did: Decentralised identifier.
    :param public_keys: Iterable containing the public keys in order of
        listing in the DID document (for anchors, starting to count from 1).
        `key_type` and `public_key` must be present for each key. Keys
        containing `authentication_type` will be registered as authentication
        keys with their given type.
    :return: DID document as a JSON string.
    """
    did_document = _DID_DOC_SKELETON.copy()
    did_document['id'] = did
    for i, pub_key in enumerate(public_keys):
        index = i + 1  # We're starting to count from 1 in the DID document.
        key_id = '{}#key{}'.format(did, index)
        key_record = {
            'id': key_id,
            'type': pub_key['key_type'],
            'controller': did,
            'publicKeyBase64': base64.b64encode(pub_key['public_key']).decode()
        }
        did_document['publicKey'].append(key_record)
        if pub_key.get('authentication_type'):
            auth_record = {
                'type': pub_key['authentication_type'],
                'publicKey': key_id
            }
            did_document['authentication'].append(auth_record)
    return json.dumps(did_document)


def api_register_did_document(did: str, did_document: str):
    """
    Register a DID document with the DID registry.

    This is an interim (centralised) solution using a centralised DID registry
    implemented via a REST API.

    :param did: Decentralised identifier.
    :param did_document: DID document to register.
    """
    resource = '{}/{}'.format(config.did_registry_resource_path, did)
    response = utils.request_api(base_url=config.did_registry_base_url,
                                 resource=resource, body=did_document,
                                 method='PUT')
    if response.status_code != 204:
        raise RuntimeError('Error registering the DID with the registry: {} {}'
                           .format(response.status_code, response.content))



# -*- coding: utf-8 -*-
"""Identity Request Service support."""

# Created: 2019-03-14 Guy K. Kloss <guy@mysinglesource.io>
#
# (c) 2018-2019 by SingleSource Limited, Auckland, New Zealand
#     http://mysinglesource.io/
#     Apache 2.0 Licence.
#
# This work is licensed under the Apache 2.0 open source licence.
# Terms and conditions apply.
#
# You should have received a copy of the licence along with this
# program.

import nacl.hash

import base58
import nacl.encoding
from sspyjose.jwk import Jwk

from singlesource.config import config  # noqa: F401 # @UnusedImport


# To load the default configuration.
_TOPIC_BYTES = 16


class EphemeralHandler:
    """
    Handler for ephemeral (single-use) communication.

    This handler is intended to be used with the ID Request Service and/or
    SingleSource messaging service.

    It provides a single-use encryption key pair for the user to receive
    confidential information with in returning messages. Additionally it
    provides a 'topic' that can be used with the SingleSource messaging
    service. This topic acts as a 'single-use post box', and is intended to
    be unguessable.
    """

    _ephemeral_jwk = None  # type: nacl.public.PrivateKey
    """Ephemeral X25519 key pair for public key content encryption."""

    def __init__(self, recovery_data: str = None):
        """
        Construct the ephemeral communication handler.

        An ephemeral handler can be de-serialised with the `recovery_data`
        parameter. This string representation can be stored elsewhere (e.g. a
        database). However, this serialised data is confidential and needs to
        be protected. It MUST NOT leak out, or the communication can be
        retrospectively decrypted by anybody who has intercepted the message
        and is in possession of this serialised recovery data.

        Note: The handler is supposed to be unguessable and for single-use
              and a particular purpose on the messaging service only. After
              completion, a new handler is to be used.

        :param recovery_data: Optional serialisation data that allows an
            ephemeral handler to be deserialised.
        """
        if recovery_data:
            self._deserialise(recovery_data)
        else:
            self._ephemeral_jwk = Jwk.get_instance(crv='X25519', generate=True)

    @property
    def messaging_topic(self) -> str:
        """
        Messaging service topic related to the ephemeral (public) key.

        Note: The topic is supposed to be unguessable and single-use for a
              particular purpose on the messaging service only.

        :return: Topic identifier.
        """
        pk_bytes = self._ephemeral_jwk.x
        hashed_pk = nacl.hash.sha256(bytes(pk_bytes),
                                     encoder=nacl.encoding.RawEncoder)
        return base58.b58encode(hashed_pk[:_TOPIC_BYTES]).decode('utf-8')

    def serialise(self) -> str:
        """
        Serialise the object to a bytes array.

        This allows an {EphemeralHandler} to be re-created after
        deserialisation.

        Note: The serialised data is confidential and needs to be protected.
              It MUST NOT leak out, or the communication can be retrospectively
              decrypted by anybody who has intercepted the message and this
              serialised handler.

        :return: Serialised content that allows the handler to be deserialised.
        """
        return base58.b58encode_check(self._ephemeral_jwk.d).decode('utf-8')

    def _deserialise(self, recovery_data: str):
        """
        Deserialise the state of an {EphemeralHandler} into this object.

        :param recovery_data: Serialised data to recover the state from.
        :raises: {ValueError} on checksum mismatch of `recovery_data`.
        """
        self._ephemeral_jwk = Jwk.get_instance(crv='X25519')
        private_key = base58.b58decode_check(recovery_data)
        self._ephemeral_jwk.d = private_key

    @property
    def private_key(self) -> bytes:
        """
        Bytes of the ephemeral private key.

        :return: Key bytes.
        """
        return self._ephemeral_jwk.d

    @property
    def public_key(self) -> bytes:
        """
        Bytes of the ephemeral public key.

        :return: Key bytes.
        """
        return self._ephemeral_jwk.x

    @property
    def jwk(self) -> Jwk:
        """
        JWK object of the ephemeral public key.

        :return: JWK.
        """
        return self._ephemeral_jwk



# -*- coding: utf-8 -*-
"""Attestation support."""

# Created: 2019-03-28 Guy K. Kloss <guy@mysinglesource.io>
#
# (c) 2018-2019 by SingleSource Limited, Auckland, New Zealand
#     http://mysinglesource.io/
#     Apache 2.0 Licence.
#
# This work is licensed under the Apache 2.0 open source licence.
# Terms and conditions apply.
#
# You should have received a copy of the licence along with this
# program.

import json
from typing import Optional, Iterable, Union, List, Dict

from pykauriid.attestations import (AttesterData,  # noqa: F401
                                    AttestationData,
                                    Attestation,
                                    AttestationStatement)
from pykauriid.claims import (ClaimSet,
                              ClaimSetKeys)
from sspyjose.jwk import Jwk

from singlesource import utils
from singlesource.config import config  # noqa: F401 # @UnusedImport


def load_attestation(attestation_content: Union[str, bytes],
                     claimsetkeys_content: Union[str, bytes],
                     attester_sig_key: Jwk,
                     subject_sig_key: Jwk) -> Attestation:
    """
    Load an attestation together with claim set keys from raw content.

    :param attestation_content: Content of the attestation to load.
    :param claimsetkeys_content: Content of the claim set keys and claim set.
    :param attester_sig_key: Key to use to verify the attestation (public).
    :param subject_sig_key: Key to use for co-signing (private).
    :return: New attestation object from data given.
    """
    an_attestation = Attestation(claimsetkeys_content,
                                 subject_signing_key=subject_sig_key,
                                 attester_signing_key=attester_sig_key)
    an_attestation.load(attestation_content)
    return an_attestation


def load_device_attestation(attestation_content: Union[str, bytes],
                            attester_sig_key: Jwk,
                            subject_sig_key: Jwk) -> Attestation:
    """
    Load an unencrypted device attestation from raw content.

    :param attestation_content: Content of the attestation to load.
    :param attester_sig_key: Key to use to verify the attestation (public).
    :param subject_sig_key: Key to use for co-signing (private).
    :return: New attestation object from data given.
    """
    an_attestation = Attestation(subject_signing_key=subject_sig_key,
                                 attester_signing_key=attester_sig_key,
                                 allow_unencrypted=True)
    an_attestation.load(attestation_content)
    return an_attestation


def make_foreign_claim_set(claims: Iterable[dict],
                           subject: str,
                           *,
                           device: bool = False) -> ClaimSetKeys:
    """
    Create a new foreign claim set on the provided claims.

    :param claims: Iterable containing JSON-LD styled claims.
    :param subject: DID of the subject of the claims.
    :param device: Set to True, if the claim set is for a device
        (default: False). Device claim sets are unencrypted.
    :return: Claim set keys object containing the foreign claim set.
    """
    claim_set = ClaimSet(sub=subject, unencrypted=device)
    claims_keys = ClaimSetKeys()
    claims_keys.claim_set = claim_set
    for claim in claims:
        claims_keys.add_claim(claim)
    claims_keys.finalise_claim_set(include_commitment=False)
    return claims_keys


def attest_claim_set(claim_set_keys: Optional[ClaimSetKeys],
                     attester_data: AttesterData,
                     attestation_data: AttestationData,
                     attester_sig_key: Union[Jwk, str, dict],
                     *,
                     claim_set: Optional[ClaimSet] = None,
                     device: bool = False) -> Dict[str, Union[str, dict]]:
    """
    Attest a (self or foreign) claim set.

    :param claim_set_keys: Claim set keys object (will be updated for the new
        trace key if required). Not required if an unencrypted (device)
        attestation is to be created.
    :param attester_data: Information on the attester.
    :param attestation_data: Data on the specific attestation to make.
    :param attester_sig_key: Key to use to attest (private).
    :param claim_set: Optional unencrypted claim set to use with devices
        (not requiring claim set keys).
    :param device: Set to True, if the attestation is for a device
        (default: False). Device attestations and claim sets are unencrypted.
    :return: Dictionary containing the attestation `id`, the serialised
        `attestation` object and updated `claim_set_keys` dictionary.
    """
    attester_sig_key = utils.jwk_from(attester_sig_key)
    my_attestation = Attestation(claim_set_keys=claim_set_keys,
                                 attester_signing_key=attester_sig_key,
                                 allow_unencrypted=device)
    if claim_set:
        my_attestation.claim_set_keys.claim_set = claim_set
    # Add attester data.
    my_attestation.attester_data = attester_data
    my_attestation.attestation_data = attestation_data
    # Attest the claim set.
    serialised_attestation = my_attestation.finalise(attester_sig_key)
    # Return attestation with updated key data (containing trace key).
    claim_set_keys_dict = (None if device else
                           my_attestation.claim_set_keys.to_dict(
                               claim_type_hints=True))
    return {
        'id': my_attestation.id,
        'attestation': serialised_attestation,
        'claim_set_keys': claim_set_keys_dict
    }


def api_request_kym_attestation(subject: str,
                                claims: dict,
                                attester_obo: str,
                                statements: List[dict]) -> Dict[str, str]:
    """
    Request a device attestation via the KYM API.

    This REST API attests a device identity *on behalf of* the requestor.

    :param subject: DID of the claim set subject.
    :param claims: Identity attribute claims to attest.
    :param attester_obo: DID of the organisation/identity on whose behalf
        the attestation will be requested.
    :param statements: List of attestation statements (JSON serialisable).
    :return: Dictionary containing the API's attestation result. It contains
        an `attestation_id` and the `attestation` itself.
    """
    request_body = {
        'subject': subject,
        'attesterObo': attester_obo,
        'claims': claims,
        'statements': statements
    }
    response = utils.request_api(base_url=config.kym_attestation_base_url,
                                 resource=config.kym_attestation_resource_path,
                                 body=json.dumps(request_body, indent=2),
                                 method='POST')
    if response.status_code != 201:
        raise RuntimeError('Error requesting a KYM attestation: {} {}'
                           .format(response.status_code, response.content))
    response_body = json.loads(response.content.decode())
    result = {
        'attestation_id': response_body['attestationId'],
        'attestation': response_body['attestation']
    }
    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Python package setup tooling."""

from codecs import open
from os import path
from setuptools.command.develop import develop
from subprocess import check_call

from setuptools import setup, find_packages


__version__ = '0.1.0'

here = path.abspath(path.dirname(__file__))


class PostDevelopCommand(develop):
    """Post-installation for development mode."""

    def run(self):
        """
        Install pre-commit git hooks for given project.
        """
        develop.run(self)
        # Install the pre-commit and pre-push hooks using `pre-commit`.
        check_call('pre-commit install'.split())
        check_call('pre-commit install --hook-type pre-push'.split())


def long_description():
    """Get the long description from the README file."""
    with open(path.join(here, 'README.md'), encoding='utf-8') as f:
        return f.read()


def install_requires(label=None, include_repo_links=True):
    """Get the dependencies and installs."""
    req_file = ('requirements-{}.txt'.format(label) if label
                else 'requirements.txt')
    with open(path.join(here, req_file), encoding='utf-8') as f:
        all_reqs = f.read().split('\n')
    packages = [x.strip() for x in all_reqs if 'git+' not in x]
    repo_links = [x.strip().replace('git+', '')
                  for x in all_reqs if x.startswith('git+')]
    if include_repo_links:
        return packages + repo_links
    else:
        return packages


setup(
    name='pysinglesource',
    version=__version__,
    description=('The high level software development kit (SDK) to'
                 ' participate in the decentralised SingleSource identity'
                 ' eco-system in Python.'),
    long_description=long_description(),
    long_description_content_type='text/markdown',
    url='https://bitbucket.org/mysinglesource/pysinglesource.git',
    license='Apache License 2.0',
    # See here for `classifiers`:
    # https://pypi.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Utilities',
        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    keywords=('singlesource blockchain data kauri kauriid identity'
              ' self-sovereign jose jwt jws jwe jwk'),
    packages=find_packages(exclude=['docs', 'tests*']),
    include_package_data=True,
    author='Guy K. Kloss',
    install_requires=install_requires(),
    dependency_links=install_requires(include_repo_links=True),
    author_email='guy@mysinglesource.io',

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'dev': install_requires('dev'),
        'test': install_requires('test')
    }
)



# -*- coding: utf-8 -*-
"""Utilities collection."""

# Created: 2019-05-08 Guy K. Kloss <guy@mysinglesource.io>
#
# (c) 2018-2019 by SingleSource Limited, Auckland, New Zealand
#     http://mysinglesource.io/
#     Apache 2.0 Licence.
#
# This work is licensed under the Apache 2.0 open source licence.
# Terms and conditions apply.
#
# You should have received a copy of the licence along with this
# program.

import base64
import json
import logging
import requests
import time
from typing import Optional, Dict, Union
import urllib

from sspyjose.jwe import Jwe
from sspyjose.jwk import Jwk

from singlesource.config import config  # noqa: F401 # @UnusedImport

# Used to cache the bearer token (access_token, expiry_ts).
_bearer_token = None  # type: Optional[tuple[str, int]]
# Used to cache the HTTP session to various hosts {host: session}.
_http_sessions = {}  # type: Dict[str, requests.Session]


def _get_http_session(url: str) -> requests.Session:
    """
    Get a session object for a particular server/host.

    :param url: URL to determine the host to cache a session object for.
    """
    host = urllib.parse.urlsplit(url).netloc
    global _http_sessions
    if host not in _http_sessions:
        logging.debug('Making new HTTP(S) session.')
        _http_sessions[host] = requests.Session()
    return _http_sessions[host]


def _make_request(url: str,
                  method: str = 'GET',
                  body: str = None,
                  *,
                  authorisation: Optional[str] = None,
                  accept: str = 'application/json',
                  headers: Optional[dict] = None) -> requests.Request:
    """
    Make an actual HTTP request to the server.

    :param url: URL to request.
    :param method: HTTP request operation (currently only "GET" and "POST").
    :param body: HTTP request body (for POST and PUT requests).
    :param authorisation: Content of HTTP `Authorization` header
        (e.g. `Bearer xyz...`).
    :param accept: Content type specification to accept (in the "Accept"
        header). Defaulting to "application/json".
    :param headers: Dictionary of additional HTTP headers to pass
        (default: None).
    :return: Request object containing the result of the request (incl.
        response). This will be the one generated by the `requests` module
        in raw.
    """
    logging.debug('Making HTTP(S) request.')
    session = _get_http_session(url)
    if not headers:
        headers = {}
    headers.update({
        'Accept': accept,
        'Accept-Encoding': ''
    })
    if authorisation:
        headers['Authorization'] = authorisation
    request = None
    if method == 'GET':
        request = session.get(url, headers=headers)
    elif method == 'POST':
        if not body:
            headers['Content-Length'] = '0'
        request = session.post(url, headers=headers, data=body)
    elif method == 'PUT':
        if not body:
            headers['Content-Length'] = '0'
        request = session.put(url, headers=headers, data=body)
    elif method == 'DELETE':
        request = session.delete(url, headers=headers)
    else:
        raise ValueError('HTTP method not supported: {}'.format(method))
    return request


def _get_oauth_token() -> Optional[str]:
    """
    Get an OAuth bearer token from the API gateway.

    :return: Valid bearer token, `None` if none is required.
    """
    if not config.oauth_consumer_key:
        # Short cut, we don't need one.
        return None

    global _bearer_token
    if _bearer_token and time.time() < _bearer_token[1]:
        return _bearer_token[0]
    host = urllib.parse.urlsplit(config.oauth_url).netloc
    auth_content = '{}:{}'.format(config.oauth_consumer_key,
                                  config.oauth_consumer_secret)
    auth_string = base64.standard_b64encode(auth_content.encode()).decode()
    authorisation = 'Basic {}'.format(auth_string)
    response = _make_request(config.oauth_url, authorisation, method='POST')
    if not response.status_code == 200:
        logging.error('Error on request {}: {} - {}'
                      .format(config.oauth_url, response.status_code,
                              response.content))
        raise RuntimeError('Could not obtain OAuth bearer token.')
    body = json.loads(response.content.decode())
    access_token = body['access_token']
    exp = int(body['issued_at']) / 1000 + int(body['expires_in'])
    _bearer_token = (access_token, exp)
    logging.debug('OAuth - host: {}, bearer token: {}, expires: {}'
                  .format(host, access_token, exp))
    return access_token


def request_api(base_url: str, resource: str, body: Optional[str] = None,
                method: str = 'GET',
                headers: Optional[dict] = None) -> requests.Request:
    """
    Make a request to the API.

    :param base_url: URL to request.
    :param resource: Resource path for the request.
    :param body: HTTP request body (for POST and PUT requests).
    :param method: HTTP request operation (currently only "GET" and "POST").
    :param headers: Dictionary of additional HTTP headers to pass
        (default: None).
    :return: Request object containing the result of the request (incl.
        response). This will be the one generated by the `requests` module
        in raw.
    """
    oauth_token = _get_oauth_token()
    authorisation = 'Bearer {}'.format(oauth_token) if oauth_token else None
    url = '{}{}'.format(base_url, resource)
    return _make_request(url, method=method, body=body,
                         authorisation=authorisation, headers=headers)


def jwk_from(jwk_data: Union[Jwk, str, dict]) -> Jwk:
    """
    Unifies a JWK key representation into a Jwk object.

    :param jwk_data: Input key data.
    :return: Jwk object from given JWK data.
    """
    if isinstance(jwk_data, str):
        jwk_data = Jwk.get_instance(from_json=jwk_data)
    elif isinstance(jwk_data, dict):
        jwk_data = Jwk.get_instance(from_dict=jwk_data)
    return jwk_data


def encrypt_to(data: dict, recipient_pub_key: Union[Jwk, str, dict]) -> str:
    """
    Encrypt data to a recipient.

    :param data: Payload data.
    :param recipient_pub_key: Public encryption key of the recipient (X25519).
    :return: JWE encrypted, string serialised cipher text.
    """
    recipient_pub_key = jwk_from(recipient_pub_key)
    encrypter = Jwe.get_instance(alg='ECDH-ES', jwk=recipient_pub_key)
    encrypter.message = data
    encrypter.encrypt()
    return encrypter.serialise()


def decrypt_from(ciphertext: Union[str, bytes],
                 own_priv_key: Union[Jwk, str, dict]) -> dict:
    """
    Decrypt data from a sender.

    :param ciphertext: JWE encrypted ciphertext in compact serialisation.
    :param own_priv_key: Own private encryption key (X25519).
    :return: Decrypted payload in dictionary form.
    :raises: An exception (depending on what cipher suite is used) in case
        of decryption or verification errors.
    """
    own_priv_key = jwk_from(own_priv_key)
    decrypter = Jwe.get_instance(alg='ECDH-ES', jwk=own_priv_key,
                                 from_compact=ciphertext)
    return decrypter.decrypt()
